MODULE TZOnce;

        (********************************************************)
        (*                                                      *)
        (*       Set OS/2's internal timezone to a value        *)
        (*      derived from the TZ environment variable.       *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            26 June 1998                    *)
        (*  Last edited:        23 April 2015                   *)
        (*  Status:             OK                              *)
        (*                                                      *)
        (********************************************************)


FROM SYSTEM IMPORT CAST, ADR;

FROM OS2 IMPORT
    (* type *)  PCSZ, DATETIME,
    (* proc *)  DosScanEnv, DosGetDateTime, DosSetDateTime;

FROM STextIO IMPORT
    (* proc *)  WriteChar, WriteString, WriteLn;

(************************************************************************)

TYPE
    CharSet = SET OF CHAR;
    MonthData = ARRAY [1..13] OF CARDINAL;
    DayOfWeek = [0..6];       (* 0 = Sunday *)

    (* Pointer to an environment string. *)

    TZPointer = POINTER TO ARRAY [0..63] OF CHAR;

    (* Our internal representation of dates and times. *)

    IDateTime = RECORD
                    year, day, minute: CARDINAL;
                    LeapYear: BOOLEAN;
                    weekday: DayOfWeek;
                END (*RECORD*);

(************************************************************************)

CONST
    (* Days since beginning of year, for the 1st of each month.  In a   *)
    (* leap year you need an extra correction.                          *)

    FirstDayInMonth = MonthData {  0,  31,  59,  90, 120, 151,
                                 181, 212, 243, 273, 304, 334, 365};

    (* Summer time rules for USA time zones.  Summer time starts at     *)
    (* 2 am on the first Sunday of April, and ends at 2 am on the last  *)
    (* Sunday of October.  If the TZ string does not include the summer *)
    (* time start/end parameters, then by convention the US rules are   *)
    (* used.  Note: this means that there is no direct way to specify   *)
    (* that there is no daylight saving period.  For that case, the TZ  *)
    (* must specify that summer time starts and ends at the same time.  *)

    USRules = "4,1,0,7200,10,-1,0,7200,3600";

(************************************************************************)

VAR
    (* A pointer to the value of the TZ environment variable.   *)

    pTZ: TZPointer;

    (* When summer time starts and ends this year. *)

    SummerTimeStart, SummerTimeEnd: IDateTime;

    (* The current date and time. *)

    Now: IDateTime;

(************************************************************************)
(*                          SCREEN OUTPUT                               *)
(************************************************************************)

PROCEDURE WriteCard1 (value: CARDINAL);

    (* Writes a one-digit number in decimal. *)

    BEGIN
        WriteChar (CHR(ORD('0') + value));
    END WriteCard1;

(************************************************************************)

PROCEDURE WriteCard2 (value: CARDINAL);

    (* Writes a two-digit number in decimal. *)

    BEGIN
        WriteCard1 (value DIV 10);
        WriteCard1 (value MOD 10);
    END WriteCard2;

(************************************************************************)

PROCEDURE WriteCard (value: CARDINAL);

    (* Writes out the number in decimal. *)

    BEGIN
        IF value > 9 THEN
            WriteCard (value DIV 10);
        END (*IF*);
        WriteCard1 (value MOD 10);
    END WriteCard;

(************************************************************************)
(*           DATES AND TIMES IN OUR INTERNAL REPRESENTATION             *)
(************************************************************************)

PROCEDURE SetDayOfWeek (VAR (*INOUT*) d: IDateTime);

    (* Calculates d.weekday, on the assumption that d.year and d.day    *)
    (* are already set correctly.                                       *)

    CONST BaseDay = 6;

    VAR FirstDayOfYear: DayOfWeek;

    BEGIN
        (* Every group of four years has 4*365+1 = 1461 days, and       *)
        (* 1461 MOD 7 = 5.  This means that the DayOfWeek changes by    *)
        (* 5 days per 4 years.                                          *)

        FirstDayOfYear := (BaseDay + 5*(d.year DIV 4)) MOD 7;

        (* Thereafter, it changes by 2 days in the first year, and one  *)
        (* day per year after that.                                     *)

        IF d.year MOD 4 <> 0 THEN
            FirstDayOfYear := (FirstDayOfYear + (d.year MOD 4) + 1) MOD 7;
        END (*IF*);

        d.weekday := (FirstDayOfYear + d.day - 1) MOD 7;

    END SetDayOfWeek;

(************************************************************************)

PROCEDURE DTtoIDT (dt: DATETIME): IDateTime;

    (* Converts from the OS/2 representation to our internal representation. *)

    VAR result: IDateTime;

    BEGIN
        result.year := dt.year;
        result.day := FirstDayInMonth[ORD(dt.month)] + ORD(dt.day) - 1;
        result.minute := 60*ORD(dt.hours) + ORD(dt.minutes);
        result.LeapYear := (dt.year MOD 4) = 0;
        IF result.LeapYear AND (ORD(dt.month) > 2) THEN
            INC (result.day);
        END (*IF*);
        result.weekday := ORD(dt.weekday);
        RETURN result;
    END DTtoIDT;

(************************************************************************)

PROCEDURE IsEqual (d1, d2: IDateTime): BOOLEAN;

    (* Returns TRUE iff d1 and d2 have the same year/day/minute values. *)

    BEGIN
        RETURN (d1.year = d2.year) AND (d1.day = d2.day)
                        AND (d1.minute = d2.minute);
    END IsEqual;

(************************************************************************)

PROCEDURE IsLater (d1, d2: IDateTime): BOOLEAN;

    (* Returns TRUE iff d1 > d2. *)

    VAR result: BOOLEAN;

    BEGIN
        IF d1.year > d2.year THEN result := TRUE
        ELSIF d1.year < d2.year THEN result := FALSE
        ELSIF d1.day > d2.day THEN result := TRUE
        ELSIF d1.day < d2.day THEN result := FALSE
        ELSIF d1.minute > d2.minute THEN result := TRUE
        ELSE result := FALSE;
        END (*IF*);
        RETURN result;
    END IsLater;

(************************************************************************)

PROCEDURE WriteDay (day: DayOfWeek);

    (* Writes the name of a day of the week. *)

    BEGIN
        CASE day OF
          |  0:  WriteString ("Sunday");
          |  1:  WriteString ("Monday");
          |  2:  WriteString ("Tuesday");
          |  3:  WriteString ("Wednesday");
          |  4:  WriteString ("Thursday");
          |  5:  WriteString ("Friday");
          |  6:  WriteString ("Saturday");
        END (*CASE*);
    END WriteDay;

(************************************************************************)

PROCEDURE WriteDate (d: IDateTime);

    (* Writes the date to the screen. *)

    BEGIN
        WriteDay (d.weekday);
        WriteString (", year ");  WriteCard (d.year);
        WriteString (", day ");  WriteCard (d.day);
        WriteString (", minute ");  WriteCard (d.minute);
    END WriteDate;

(************************************************************************)
(*               ANALYSIS OF THE TZ ENVIRONMENT VARIABLE                *)
(************************************************************************)

PROCEDURE LoadTZvar;

    (* Reads the TZ environment variable, gives a value to the global pTZ. *)

    VAR pszValue: PCSZ;

    BEGIN
        WriteString ("The value of the TZ environment variable is");
        IF DosScanEnv ("TZ", pszValue) <> 0 THEN
            WriteString (" undefined");
            pTZ := NIL;
        ELSE
            WriteLn;  WriteString ("     ");
            pTZ := CAST (TZPointer, pszValue);
            WriteString (pTZ^);
        END (*IF*);
        WriteLn;
    END LoadTZvar;

(************************************************************************)

PROCEDURE AnalyseTZvar (VAR (*OUT*) BaseZone, correction: INTEGER);

    (* Extracts information from the TZ string. *)

    VAR j: CARDINAL;

    (********************************************************************)

    PROCEDURE ReadSign(): BOOLEAN;

        (* Reads '+' or '-' from pTZ^[j], updates j, returns TRUE       *)
        (* if we found '-'.  (The '+' is optional.)                     *)

        VAR negative: BOOLEAN;

        BEGIN
            negative := FALSE;
            IF pTZ^[j] = '+' THEN
                INC (j);
            ELSIF pTZ^[j] = '-' THEN
                negative := TRUE;  INC (j);
            END (*IF*);
            RETURN negative;
        END ReadSign;

    (********************************************************************)

    PROCEDURE ReadNum(): INTEGER;

        (* Reads a numeric string from pTZ^[j], updates j. *)

        VAR result: INTEGER;  negative: BOOLEAN;

        BEGIN
            negative := ReadSign();
            result := 0;
            WHILE pTZ^[j] IN CharSet {'0'..'9'} DO
                result := 10*result + VAL(INTEGER, ORD(pTZ^[j]) - ORD('0'));
                INC (j);
            END (*WHILE*);

            IF negative THEN
                result := -result;
            END (*IF*);
            RETURN result;

        END ReadNum;

    (********************************************************************)

    PROCEDURE DecodeBoundary (VAR (*OUT*) result: IDateTime);

        (* Converts a four-parameter string, used in TZ to specify the  *)
        (* start or end of summer time, to IDateTime format.            *)

        VAR num: ARRAY [0..3] OF INTEGER;  k: CARDINAL;
            TargetDay: DayOfWeek;

        BEGIN
            (* Copy the "Now" variable, for the sake of the year. *)

            result := Now;

            (* Gets the four numbers from the TZ string. *)

            FOR k := 0 TO 3 DO
                IF pTZ^[j] = ',' THEN
                    INC (j);
                END (*IF*);
                num[k] := ReadNum();
            END (*FOR*);

            (* The first of the four numbers is a month. *)

            result.day := FirstDayInMonth[num[0]];
            IF result.LeapYear AND (num[0] > 2) THEN
                INC (result.day);
            END (*IF*);

            (* The second and third numbers give the day in the month,  *)
            (* but it's in a non-obvious format.  The third number is a *)
            (* day of the week (0 = Sunday).  The second, if it's       *)
            (* positive, says whether it's the first, second, etc.      *)
            (* occurrence of that day in the month.  If it's negative,  *)
            (* we count backwards from the end of the month.  If it's   *)
            (* zero, then the third number is an actual day of the      *)
            (* month, i.e. it's not a day-within-week.                  *)

            IF num[1] = 0 THEN
                INC (result.day, num[2] - 1);
                SetDayOfWeek (result);
            ELSE
                SetDayOfWeek (result);
                TargetDay := ABS(num[2]);
                IF num[1] > 0 THEN

                    (* Counting from first day of month. *)

                    WHILE result.weekday <> TargetDay DO
                        INC (result.day);
                        result.weekday := (result.weekday + 1) MOD 7;
                    END (*WHILE*);
                    DEC (num[1]);
                    INC (result.day, 7*ABS(num[1]));
                ELSE

                    (* Counting from last day of month. *)

                    INC (result.day, FirstDayInMonth[num[0]+1]
                               - FirstDayInMonth[num[0]] - 1);
                    IF result.LeapYear AND (num[0] = 2) THEN
                        INC (result.day);
                    END (*IF*);
                    SetDayOfWeek (result);

                    WHILE result.weekday <> TargetDay DO
                        DEC (result.day);
                        IF result.weekday = 0 THEN result.weekday := 6
                        ELSE DEC(result.weekday)
                        END (*IF*);
                    END (*WHILE*);
                    INC (num[1]);
                    DEC (result.day, 7*ABS(num[1]));

                END (*IF*);
            END (*IF*);
            result.minute := (num[3] + 30) DIV 60;

        END DecodeBoundary;

    (********************************************************************)

    VAR negative: BOOLEAN;
        txtbuffer: ARRAY [0..31] OF CHAR;

    BEGIN
        j := 0;

        (* The first parameter is AAAnBBB, where AAA and BBB are textual codes  *)
        (* of no interest to us, and n is a signed hour:minute value where the  *)
        (* colon and minute parts are optional.  This gives the time zone when  *)
        (* summer time is not in force.  The value is negative for regions      *)
        (* east of Greenwich.                                                   *)

        WHILE NOT (pTZ^[j] IN CharSet {'+', '-', '0'..'9'}) DO
            INC (j);
        END (*WHILE*);

        negative := ReadSign();
        BaseZone := 60*ReadNum();
        IF pTZ^[j] = ':' THEN
            INC (j);  INC (BaseZone, ReadNum());
        END (*IF*);
        IF negative THEN
            BaseZone := -BaseZone;
        END (*IF*);

        WHILE (pTZ^[j] <> CHR(0)) AND (pTZ^[j] <> ',') DO
            INC (j);
        END (*WHILE*);

        (* If we've reached the end of the string, the default summer time  *)
        (* information must be appended.                                    *)

        IF pTZ^[j] = CHR(0) THEN
            WriteString ("No summer time corrections included in TZ.");
            WriteLn;
            WriteString ("Using North American rules by default.");
            WriteLn;
            txtbuffer := USRules;  pTZ := ADR(txtbuffer);  j := 0;
        END (*IF*);

        DecodeBoundary (SummerTimeStart);
        DecodeBoundary (SummerTimeEnd);

        (* Finally we should have a comma followed by the summer time       *)
        (* correction in seconds.  Convert this to minutes.                 *)

        IF pTZ^[j] = ',' THEN
            INC (j);
        END (*IF*);
        correction := -((ReadNum() + 30) DIV 60);

    END AnalyseTZvar;

(********************************************************************************)
(*                       THE TIME ZONE AS STORED BY OS/2                        *)
(********************************************************************************)

PROCEDURE WriteTimeZone (value: INTEGER);

    (* Takes a value in minutes, writes it in +hhmm format.  Note that the      *)
    (* value we report here is actually the negative of the value that is used  *)
    (* by DosGetDateTime or by the TZ environment variable.                     *)

    BEGIN
        IF value < 0 THEN
            WriteChar ('+');  value := -value;
        ELSE
            WriteChar ('-');
        END (*IF*);
        WriteCard2 (value DIV 60);
        WriteCard2 (value MOD 60);
    END WriteTimeZone;

(********************************************************************************)

PROCEDURE ReportTimeZone;

    (* Rereads the clock, reports what time zone OS/2 has stored. *)
    (* Also gives a value to the global variable "Now".           *)

    VAR dt: DATETIME;

    BEGIN
        DosGetDateTime (dt);
        Now := DTtoIDT (dt);
        WriteString ("The current time zone setting is ");
        IF dt.timezone = -1 THEN
            WriteString ("undefined");
        ELSE
            WriteTimeZone (dt.timezone);
        END (*IF*);
        WriteLn;
    END ReportTimeZone;

(********************************************************************************)

PROCEDURE UpdateTimeZone (value: INTEGER);

    (* Stores an updated value in the system's copy of the time zone. *)

    VAR dt: DATETIME;

    BEGIN
        DosGetDateTime (dt);
        dt.timezone := value;
        DosSetDateTime (dt);
    END UpdateTimeZone;

(********************************************************************************)
(*                              MAIN PROGRAM                                    *)
(********************************************************************************)

PROCEDURE MainProg;

    VAR SummerHasStarted, SummerIsNotOver, WeAreInSummerTime: BOOLEAN;
        BaseTimeZone, correction:INTEGER;

    BEGIN
        WriteLn;
        ReportTimeZone;
        WriteString ("It is now ");  WriteDate (Now);  WriteLn;
        LoadTZvar;
        AnalyseTZvar (BaseTimeZone, correction);
        WriteString ("Your base time zone is ");
        WriteTimeZone (BaseTimeZone);  WriteLn;

        IF IsEqual (SummerTimeStart, SummerTimeEnd) THEN
            WriteString ("Summer time adjustments are disabled.");
            WriteLn;
            correction := 0;
        ELSE

            SummerHasStarted := IsLater (Now, SummerTimeStart);
            SummerIsNotOver := IsLater (SummerTimeEnd, Now);

            IF IsLater (SummerTimeStart, SummerTimeEnd) THEN

                (* Southern hemisphere *)

                WeAreInSummerTime := SummerHasStarted OR SummerIsNotOver;

            ELSE

                (* Northern hemisphere *)

                WeAreInSummerTime := SummerHasStarted AND SummerIsNotOver;

            END (*IF*);

            WriteString ("You are now on ");
            IF WeAreInSummerTime THEN
                WriteString ("summer");
            ELSE
                WriteString ("winter");
            END (*IF*);
            WriteString (" time");  WriteLn;

            WriteString ("Summer time start");
            IF SummerHasStarted THEN
                WriteString ("ed");
            ELSE
                WriteChar ('s');
            END (*IF*);
            WriteString (" on ");
            WriteDate (SummerTimeStart);  WriteLn;

            WriteString ("Summer time end");
            IF SummerIsNotOver THEN
                WriteChar ('s');
            ELSE
                WriteString ("ed");
            END (*IF*);
            WriteString (" on ");
            WriteDate (SummerTimeEnd);  WriteLn;

            (* Report the summer time correction, if applicable. *)

            IF WeAreInSummerTime THEN
                WriteString ("Time zone will be adjusted to ");
                WriteTimeZone (BaseTimeZone + correction);  WriteLn;
            ELSE
                correction := 0;
            END (*IF*);

        END (*IF*);

        UpdateTimeZone (BaseTimeZone + correction);
        ReportTimeZone;

    END MainProg;

(********************************************************************************)

BEGIN
    MainProg;
END TZOnce.

