MODULE TZSet;

        (********************************************************)
        (*                                                      *)
        (*       Set OS/2's internal timezone to a value        *)
        (*      derived from the TZ environment variable.       *)
        (*                                                      *)
        (*   Most of the time this program is doing nothing,    *)
        (*   but it wakes up every so often to see whether a    *)
        (*   change in time zone offset is needed.              *)
        (*                                                      *)
        (*   NOTE: the leap year calculations in this module    *)
        (*   are valid only until the year 2099.                *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Last edited:        16 April 2014                   *)
        (*  Status:             Working                         *)
        (*                                                      *)
        (********************************************************)


FROM SYSTEM IMPORT CAST, ADR, LOC, ADDRESS;

IMPORT OS2;

<* IF SCREENOUT THEN *>
FROM STextIO IMPORT
    (* proc *)  WriteChar, WriteString, WriteLn;
<* END *>

(************************************************************************)

CONST version = "06";

(************************************************************************)
(*                          TYPES AND CONSTANTS                         *)
(************************************************************************)

TYPE
    CharSet = SET OF CHAR;
    MonthData = ARRAY [1..13] OF CARDINAL;
    DayOfWeek = [0..6];       (* 0 = Sunday *)

    (* Pointer to an environment string. *)

    TZPointer = POINTER TO ARRAY [0..63] OF CHAR;

    (* Our internal representation of dates and times.  For our purpose *)
    (* we only need one-minute resolution, and we do not need the       *)
    (* concepts of months or hours except when converting to/from more  *)
    (* conventional notation.                                           *)

    IDateTime = RECORD
                    year, day, minute: CARDINAL;
                    weekday: DayOfWeek;
                END (*RECORD*);

(************************************************************************)

CONST
    STACKSIZE = 8192;

    (* Days since beginning of year, for the 1st of each month.  In a   *)
    (* leap year you need an extra correction.                          *)

    FirstDayInMonth = MonthData {  0,  31,  59,  90, 120, 151,
                                 181, 212, 243, 273, 304, 334, 365};

    Feb28 = 31 + 28;

    (* Summer time rules for USA time zones.  Summer time starts at     *)
    (* 2 am on the first Sunday of April, and ends at 2 am on the last  *)
    (* Sunday of October.  If the TZ string does not include the summer *)
    (* time start/end parameters, then by convention the US rules are   *)
    (* used.  Note: this means that there is no direct way to specify   *)
    (* that there is no daylight saving period.  For that case, the TZ  *)
    (* must specify that summer time starts and ends at the same time.  *)

    USRules = "4,1,0,7200,10,-1,0,7200,3600";

(************************************************************************)
(*                           GLOBAL VARIABLES                           *)
(************************************************************************)

VAR
    (* When summer time next starts and ends. *)

    SummerTimeStart, SummerTimeEnd: IDateTime;

    (* Time zone offset from GMT, and correction to apply in summer.    *)
    (* Both of these values are in minutes.                             *)

    BaseTimeZone, SummerAdjust: INTEGER;

    (* An event semaphore that controls the timing of the main thread.  *)

    MainSem: OS2.HEV;

    (* Are we currently in summer? *)

    WeAreInSummerTime: BOOLEAN;

    (* A flag to tell the main thread to exit. *)

    Shutdown: BOOLEAN;

(************************************************************************)
(*                          SCREEN OUTPUT                               *)
(************************************************************************)

<* IF SCREENOUT THEN *>

PROCEDURE WriteCard1 (value: CARDINAL);

    (* Writes a one-digit number in decimal. *)

    BEGIN
        WriteChar (CHR(ORD('0') + value));
    END WriteCard1;

(************************************************************************)

PROCEDURE WriteCard2 (value: CARDINAL);

    (* Writes a two-digit number in decimal. *)

    BEGIN
        WriteCard1 (value DIV 10);
        WriteCard1 (value MOD 10);
    END WriteCard2;

(************************************************************************)

PROCEDURE WriteCard (value: CARDINAL);

    (* Writes out the number in decimal. *)

    BEGIN
        IF value > 9 THEN
            WriteCard (value DIV 10);
        END (*IF*);
        WriteCard1 (value MOD 10);
    END WriteCard;

<* END *>

(************************************************************************)
(*           DATES AND TIMES IN OUR INTERNAL REPRESENTATION             *)
(************************************************************************)

PROCEDURE SetDayOfWeek (VAR (*INOUT*) d: IDateTime);

    (* Calculates d.weekday, on the assumption that d.year and d.day    *)
    (* are already set correctly.                                       *)

    CONST BaseDay = 6;

    VAR FirstDayOfYear: DayOfWeek;

    BEGIN
        (* Every group of four years has 4*365+1 = 1461 days, and       *)
        (* 1461 MOD 7 = 5.  This means that the DayOfWeek changes by    *)
        (* 5 days per 4 years.                                          *)

        FirstDayOfYear := (BaseDay + 5*(d.year DIV 4)) MOD 7;

        (* Thereafter, it changes by 2 days in the first year, and one  *)
        (* day per year after that.                                     *)

        IF d.year MOD 4 <> 0 THEN
            FirstDayOfYear := (FirstDayOfYear + (d.year MOD 4) + 1) MOD 7;
        END (*IF*);

        d.weekday := (FirstDayOfYear + d.day - 1) MOD 7;

    END SetDayOfWeek;

(************************************************************************)

PROCEDURE DTtoIDT (dt: OS2.DATETIME): IDateTime;

    (* Converts from the OS/2 representation to our internal representation. *)

    VAR result: IDateTime;
        IsLeapYear: BOOLEAN;

    BEGIN
        result.year := dt.year;
        result.day := FirstDayInMonth[ORD(dt.month)] + ORD(dt.day) - 1;
        result.minute := 60*ORD(dt.hours) + ORD(dt.minutes);
        IsLeapYear := (dt.year MOD 4) = 0;
        IF IsLeapYear AND (ORD(dt.month) > 2) THEN
            INC (result.day);
        END (*IF*);
        result.weekday := ORD(dt.weekday);
        RETURN result;
    END DTtoIDT;

(************************************************************************)

PROCEDURE IsEqual (d1, d2: IDateTime): BOOLEAN;

    (* Returns TRUE iff d1 and d2 have the same year/day/minute values. *)

    BEGIN
        RETURN (d1.year = d2.year) AND (d1.day = d2.day)
                        AND (d1.minute = d2.minute);
    END IsEqual;

(************************************************************************)

PROCEDURE IsLater (d1, d2: IDateTime): BOOLEAN;

    (* Returns TRUE iff d1 > d2. *)

    VAR result: BOOLEAN;

    BEGIN
        IF d1.year > d2.year THEN result := TRUE
        ELSIF d1.year < d2.year THEN result := FALSE
        ELSIF d1.day > d2.day THEN result := TRUE
        ELSIF d1.day < d2.day THEN result := FALSE
        ELSIF d1.minute > d2.minute THEN result := TRUE
        ELSE result := FALSE;
        END (*IF*);
        RETURN result;
    END IsLater;

(************************************************************************)

<* IF SCREENOUT THEN *>

PROCEDURE WriteDate (d: IDateTime);

    (* Writes the date to the screen. *)

    BEGIN
        CASE d.weekday OF
          |  0:  WriteString ("Sunday");
          |  1:  WriteString ("Monday");
          |  2:  WriteString ("Tuesday");
          |  3:  WriteString ("Wednesday");
          |  4:  WriteString ("Thursday");
          |  5:  WriteString ("Friday");
          |  6:  WriteString ("Saturday");
        END (*CASE*);
        WriteString (", year ");  WriteCard (d.year);
        WriteString (", day ");  WriteCard (d.day);
        WriteString (", minute ");  WriteCard (d.minute);
    END WriteDate;

<* END *>

(************************************************************************)
(*                                                                      *)
(*               ANALYSIS OF THE TZ ENVIRONMENT VARIABLE                *)
(*                                                                      *)
(*  We leave results in the globals SummerTimeStart, SummerTimeEnd.     *)
(*                                                                      *)
(************************************************************************)

PROCEDURE LoadTZvar(): TZPointer;

    (* Reads the TZ environment variable, returns a pointer to it. *)

    VAR pszValue: OS2.PCSZ;
        pTZ: TZPointer;

    BEGIN
        IF OS2.DosScanEnv ("TZ", pszValue) <> 0 THEN
            pTZ := NIL;
        ELSE
            pTZ := CAST (TZPointer, pszValue);
        END (*IF*);

        <* IF SCREENOUT THEN *>
            WriteString ("The value of the TZ environment variable is");
            IF pTZ = NIL THEN
                WriteString (" undefined");
            ELSE
                WriteLn;  WriteString ("     ");
                WriteString (pTZ^);
            END (*IF*);
            WriteLn;
        <* END *>

        RETURN pTZ;

    END LoadTZvar;

(************************************************************************)

PROCEDURE AnalyseTZvar (thisyear: CARDINAL;
                              VAR (*OUT*) BaseZone, correction: INTEGER);

    (* Extracts information from the TZ string. *)

    VAR j: CARDINAL;
        pTZ: TZPointer;

    (********************************************************************)

    PROCEDURE ReadSign(): BOOLEAN;

        (* Reads '+' or '-' from pTZ^[j], updates j, returns TRUE       *)
        (* if we found '-'.  (The '+' is optional.)                     *)

        VAR negative: BOOLEAN;

        BEGIN
            negative := FALSE;
            IF pTZ^[j] = '+' THEN
                INC (j);
            ELSIF pTZ^[j] = '-' THEN
                negative := TRUE;  INC (j);
            END (*IF*);
            RETURN negative;
        END ReadSign;

    (********************************************************************)

    PROCEDURE ReadNum(): INTEGER;

        (* Reads a numeric string from pTZ^[j], updates j. *)

        VAR result: INTEGER;  negative: BOOLEAN;

        BEGIN
            negative := ReadSign();
            result := 0;
            WHILE pTZ^[j] IN CharSet {'0'..'9'} DO
                result := 10*result + VAL(INTEGER, ORD(pTZ^[j]) - ORD('0'));
                INC (j);
            END (*WHILE*);

            IF negative THEN
                result := -result;
            END (*IF*);
            RETURN result;

        END ReadNum;

    (********************************************************************)

    PROCEDURE DecodeBoundary (thisyear: CARDINAL;
                                 VAR (*OUT*) result: IDateTime);

        (* Converts a four-parameter string, used in TZ to specify the  *)
        (* start or end of summer time, to IDateTime format.            *)

        VAR num: ARRAY [0..3] OF INTEGER;  k: CARDINAL;
            TargetDay: DayOfWeek;  IsLeapYear: BOOLEAN;

        BEGIN
            (* Set the year in the result. *)

            result.year := thisyear;
            IsLeapYear := (result.year MOD 4) = 0;

            (* Gets the four numbers from the TZ string. *)

            FOR k := 0 TO 3 DO
                IF pTZ^[j] = ',' THEN
                    INC (j);
                END (*IF*);
                num[k] := ReadNum();
            END (*FOR*);

            (* The first of the four numbers is a month. *)

            result.day := FirstDayInMonth[num[0]];
            IF IsLeapYear AND (num[0] > 2) THEN
                INC (result.day);
            END (*IF*);

            (* The second and third numbers give the day in the month,  *)
            (* but it's in a non-obvious format.  The third number is a *)
            (* day of the week (0 = Sunday).  The second, if it's       *)
            (* positive, says whether it's the first, second, etc.      *)
            (* occurrence of that day in the month.  If it's negative,  *)
            (* we count backwards from the end of the month.  If it's   *)
            (* zero, then the third number is an actual day of the      *)
            (* month, i.e. it's not a day-within-week.                  *)

            IF num[1] = 0 THEN
                INC (result.day, num[2] - 1);
                SetDayOfWeek (result);
            ELSE
                SetDayOfWeek (result);
                TargetDay := ABS(num[2]);
                IF num[1] > 0 THEN

                    (* Counting from first day of month. *)

                    WHILE result.weekday <> TargetDay DO
                        INC (result.day);
                        result.weekday := (result.weekday + 1) MOD 7;
                    END (*WHILE*);
                    DEC (num[1]);
                    INC (result.day, 7*ABS(num[1]));
                ELSE

                    (* Counting from last day of month. *)

                    INC (result.day, FirstDayInMonth[num[0]+1]
                               - FirstDayInMonth[num[0]] - 1);
                    IF IsLeapYear AND (num[0] = 2) THEN
                        INC (result.day);
                    END (*IF*);
                    SetDayOfWeek (result);

                    WHILE result.weekday <> TargetDay DO
                        DEC (result.day);
                        IF result.weekday = 0 THEN result.weekday := 6
                        ELSE DEC(result.weekday)
                        END (*IF*);
                    END (*WHILE*);
                    INC (num[1]);
                    DEC (result.day, 7*ABS(num[1]));

                END (*IF*);
            END (*IF*);
            result.minute := (num[3] + 30) DIV 60;

        END DecodeBoundary;

    (********************************************************************)

    CONST DummyResult = IDateTime{0,0,0,0};

    VAR negative: BOOLEAN;
        txtbuffer: ARRAY [0..31] OF CHAR;

    BEGIN
        pTZ := LoadTZvar();

        IF pTZ = NIL THEN
            SummerTimeStart := DummyResult;
            SummerTimeEnd := DummyResult;
            BaseTimeZone := 0;  correction := 0;
            RETURN;
        END (*IF*);

        j := 0;

        (* The first parameter is AAAnBBB, where AAA and BBB are codes  *)
        (* that are of no interest to us, and n is a signed hour:minute *)
        (* value where the colon and minute parts are optional.  This   *)
        (* gives the time zone when summer time is not in force.  The   *)
        (* value is negative for regions east of Greenwich.             *)

        WHILE NOT (pTZ^[j] IN CharSet {'+', '-', '0'..'9'}) DO
            INC (j);
        END (*WHILE*);

        negative := ReadSign();
        BaseZone := 60*ReadNum();
        IF pTZ^[j] = ':' THEN
            INC (j);  INC (BaseZone, ReadNum());
        END (*IF*);
        IF negative THEN
            BaseZone := -BaseZone;
        END (*IF*);

        WHILE (pTZ^[j] <> CHR(0)) AND (pTZ^[j] <> ',') DO
            INC (j);
        END (*WHILE*);

        (* If we've reached the end of the string, the default summer   *)
        (* time information must be appended.                           *)

        IF pTZ^[j] = CHR(0) THEN
            <* IF SCREENOUT THEN *>
                WriteString ("No summer time corrections included in TZ.");
                WriteLn;
                WriteString ("Using North American rules by default.");
                WriteLn;
            <* END *>;
            txtbuffer := USRules;  pTZ := ADR(txtbuffer);  j := 0;
        END (*IF*);

        DecodeBoundary (thisyear, SummerTimeStart);
        DecodeBoundary (thisyear, SummerTimeEnd);

        (* Finally we should have a comma followed by the summer time   *)
        (* correction in seconds.  Convert this to minutes.             *)

        IF pTZ^[j] = ',' THEN
            INC (j);
        END (*IF*);
        correction := -((ReadNum() + 30) DIV 60);

    END AnalyseTZvar;

(************************************************************************)
(*                   THE TIME ZONE AS STORED BY OS/2                    *)
(************************************************************************)

<* IF SCREENOUT THEN *>

PROCEDURE WriteTimeZone (value: INTEGER);

    (* Takes a value in minutes, writes it in +hhmm format.  Note that  *)
    (* the value we report here is actually the negative of the value   *)
    (* that is used by DosGetDateTime or by the TZ environment variable.*)

    BEGIN
        IF value < 0 THEN
            WriteChar ('+');  value := -value;
        ELSE
            WriteChar ('-');
        END (*IF*);
        WriteCard2 (value DIV 60);
        WriteCard2 (value MOD 60);
    END WriteTimeZone;

<* END *>

(************************************************************************)

PROCEDURE UpdateTimeZone (value: INTEGER);

    (* Stores an updated value in the system's copy of the time zone.   *)
    (* We use two kinds of critical section protection here.  The       *)
    (* priority setting reduces the likelihood of clock changes by      *)
    (* some other program.  The "must complete" protection ensures that *)
    (* we don't get interrupted by a process shutdown.                  *)

    VAR dt: OS2.DATETIME;  rc: OS2.APIRET;  nesting: CARDINAL;

    BEGIN
        rc := OS2.DosEnterMustComplete (nesting);
        rc := OS2.DosSetPriority (OS2.PRTYS_THREAD,
                                    OS2. PRTYC_TIMECRITICAL, 0, 0);
        OS2.DosGetDateTime (dt);
        IF dt.timezone <> value THEN
            dt.timezone := value;
            OS2.DosSetDateTime (dt);
        END (*IF*);
        rc := OS2.DosSetPriority (OS2.PRTYS_THREAD,
                                    OS2. PRTYC_REGULAR, 0, 0);
        rc := OS2.DosExitMustComplete (nesting);
    END UpdateTimeZone;

(************************************************************************)
(*                           MOVING TO NEXT YEAR                        *)
(************************************************************************)

PROCEDURE AddOneYear (VAR (*INOUT*) date: IDateTime);

    (* Adds one calendar year to date. *)

    VAR leap: CARDINAL;

    BEGIN
        leap := date.year MOD 4;
        IF (leap = 0) AND (date.day > Feb28) THEN
            DEC (date.day);
        END (*IF*);
        INC (date.year);  INC(leap);
        IF (leap = 4) AND (date.day > Feb28) THEN
            INC (date.day);
        END (*IF*);
    END AddOneYear;

(************************************************************************)
(*                          MAIN CALCULATION                            *)
(************************************************************************)

PROCEDURE GetTZinfo;

    (* Sets the values of global variables SummerTimeStart,             *)
    (* SummerTimeEnd, BaseTimeZone, SummerAdjust, and WeAreInSummerTime.*)

    VAR dt: OS2.DATETIME;  Now: IDateTime;
        SummerHasStarted, SummerIsNotOver: BOOLEAN;

    BEGIN
        OS2.DosGetDateTime (dt);
        Now := DTtoIDT (dt);
        AnalyseTZvar (Now.year, BaseTimeZone, SummerAdjust);
        <* IF SCREENOUT THEN *>
            WriteString ("The current time zone setting is ");
            IF dt.timezone = -1 THEN
                WriteString ("undefined");
            ELSE
                WriteTimeZone (dt.timezone);
            END (*IF*);
            WriteLn;
            WriteString ("It is now ");  WriteDate (Now);  WriteLn;
            WriteString ("Your base time zone is ");
            WriteTimeZone (BaseTimeZone);  WriteLn;
        <* END *>

        IF IsEqual (SummerTimeStart, SummerTimeEnd) THEN
            <* IF SCREENOUT THEN *>
                WriteString ("Summer time adjustments are disabled.");
                WriteLn;
            <* END *>
            SummerAdjust := 0;
        ELSE

            SummerHasStarted := NOT IsLater (SummerTimeStart, Now);
            SummerIsNotOver := IsLater (SummerTimeEnd, Now);

            IF IsLater (SummerTimeStart, SummerTimeEnd) THEN

                (* Southern hemisphere *)

                WeAreInSummerTime := SummerHasStarted OR SummerIsNotOver;

            ELSE

                (* Northern hemisphere *)

                WeAreInSummerTime := SummerHasStarted AND SummerIsNotOver;

            END (*IF*);

            <* IF SCREENOUT THEN *>
                WriteString ("You are now on ");
                IF WeAreInSummerTime THEN
                    WriteString ("summer");
                ELSE
                    WriteString ("winter");
                END (*IF*);
                WriteString (" time");  WriteLn;

                (* Report the summer time correction, if applicable. *)

                IF WeAreInSummerTime THEN
                    WriteString ("Time zone will be set to ");
                    WriteTimeZone (BaseTimeZone + SummerAdjust);  WriteLn;
                END (*IF*);
            <* END *>

            (* Up to this point, SummerTimeStart and SummerTimeEnd are  *)
            (* dates in this year.  Correct them to be future dates.    *)

            IF SummerHasStarted THEN
                AddOneYear (SummerTimeStart);
            END (*IF*);
            IF NOT SummerIsNotOver THEN
                AddOneYear (SummerTimeEnd);
            END (*IF*);

            <* IF SCREENOUT THEN *>
                IF WeAreInSummerTime THEN
                    WriteString ("Summer time  ends  on ");
                    WriteDate (SummerTimeEnd);  WriteLn;
                END (*IF*);
                WriteString ("Summer time starts on ");
                WriteDate (SummerTimeStart);  WriteLn;
                IF NOT WeAreInSummerTime THEN
                    WriteString ("Summer time  ends  on ");
                    WriteDate (SummerTimeEnd);  WriteLn;
                END (*IF*);
            <* END *>

        END (*IF*);

    END GetTZinfo;

(************************************************************************)

PROCEDURE DiffMinutes (VAR (*IN*) from, to: IDateTime): INTEGER;

    (* Returns (to - from), in minutes.  Assumption: the dates are      *)
    (* either in the same year, or to is in the year after from.        *)

    VAR result: INTEGER;

    BEGIN
        result := 0;
        IF to.year > from.year THEN
            result := 365;

            (* Increment the result if either from is a leap year and   *)
            (* from.day is before Feb 29, or to is a leap year and      *)
            (* to.day is after Feb 28.  Because we assuming at most a   *)
            (* one-year gap, only one of these conditions can hold.     *)

            IF ((from.year MOD 4 = 0) AND (from.day <= Feb28))
                       OR ((to.year MOD 4 = 0) AND (to.day > Feb28)) THEN
                result := 366;
            END (*IF*);

        END (*IF*);
        result := result + CAST(INTEGER,to.day) - CAST(INTEGER,from.day);
        result := 60*24*result
                  + CAST(INTEGER,to.minute) - CAST(INTEGER,from.minute);
        RETURN result;
    END DiffMinutes;

(************************************************************************)

PROCEDURE TimeToNext(): INTEGER;

    (* Returns the time until the next correction is due, in minutes. *)

    VAR MinsToNext: INTEGER;
        dt: OS2.DATETIME;
        Now: IDateTime;

    BEGIN
        OS2.DosGetDateTime (dt);
        Now := DTtoIDT (dt);
        IF WeAreInSummerTime THEN
            MinsToNext := DiffMinutes (Now, SummerTimeEnd);
        ELSE
            MinsToNext := DiffMinutes (Now, SummerTimeStart);
        END (*IF*);
        <* IF SCREENOUT THEN *>
            IF MinsToNext < 0 THEN
                WriteChar ('-');
            END (*IF*);
            WriteCard (ABS(MinsToNext));
            WriteString (" minutes until next time zone change.");  WriteLn;
        <* END *>

        RETURN MinsToNext;

    END TimeToNext;

(************************************************************************)

PROCEDURE MinsToms (minutes: INTEGER): INTEGER;

    (* Converts minutes to milliseconds, with saturation if this would  *)
    (* cause arithmetic overflow.                                       *)

    VAR result: INTEGER;

    BEGIN
        IF minutes > MAX(INTEGER)/(1000*60) THEN
            result := MAX(INTEGER);
        ELSE
            result := 1000*60*minutes;

            (* Add half a second, for conservatism's sake. *)

            INC (result, 500);

        END (*IF*);

        RETURN result;

     END MinsToms;

(************************************************************************)

PROCEDURE DoSetting(): INTEGER;

    (* Sets the time zone, returns the time to the next adjustment. *)

    VAR MinsToNext: INTEGER;

    BEGIN
        MinsToNext := TimeToNext();

        (* Here is the actual correction. *)

        IF WeAreInSummerTime THEN
            UpdateTimeZone (BaseTimeZone + SummerAdjust);
        ELSE
            UpdateTimeZone (BaseTimeZone);
        END (*IF*);

        RETURN MinsToNext;

    END DoSetting;

(************************************************************************)
(*                 PROVISION TO SHUT DOWN THE PROGRAM                   *)
(************************************************************************)

PROCEDURE ["SysCall"] BreakHandler (p1: OS2.PEXCEPTIONREPORTRECORD;
                                     p2: OS2.PEXCEPTIONREGISTRATIONRECORD;
                                     p3: OS2.PCONTEXTRECORD;
                                     p4: ADDRESS): CARDINAL;

    (* This handler shuts down the program when an XCPT_SIGNAL is received.  *)

    BEGIN
        IF p1^.ExceptionNum = OS2.XCPT_SIGNAL THEN
            <* IF SCREENOUT THEN *>
                WriteString ("Shutting down.");
                WriteLn;
            <* END *>
            Shutdown := TRUE;
            OS2.DosPostEventSem (MainSem);
            RETURN OS2.XCPT_CONTINUE_EXECUTION;
        ELSE
            RETURN OS2.XCPT_CONTINUE_SEARCH;
        END (*IF*);
    END BreakHandler;

(************************************************************************)

PROCEDURE [OS2.APIENTRY] TaskKiller (dummy: CARDINAL);

    (* A separate thread, to shut down this program if requested.  The  *)
    (* goal is to prevent having two copies of the program running.     *)

    VAR semName: ARRAY [0..127] OF CHAR;
        ShutdownSignal: OS2.HEV;

    BEGIN
        IF dummy = 0 THEN END(*IF*);  (* to suppress a compiler warning *)

        semName := "\SEM32\TZSET_TERMINATE";
        ShutdownSignal := 0;
        WHILE OS2.DosOpenEventSem (semName, ShutdownSignal)
                                            <> OS2.ERROR_SEM_NOT_FOUND DO

            (* Another instance of this program must be running.  Kill  *)
            (* it, and don't proceed until we know that that other      *)
            (* instance has closed the semaphore.                       *)

            OS2.DosPostEventSem (ShutdownSignal);
            OS2.DosSleep (500);
        END (*WHILE*);

        OS2.DosCreateEventSem (semName, ShutdownSignal, 0, FALSE);

        (* Wait for a signal on the shutdown semaphore.    *)

        WHILE NOT Shutdown DO
            OS2.DosWaitEventSem (ShutdownSignal, OS2.SEM_INDEFINITE_WAIT);

            <* IF SCREENOUT THEN *>
                WriteString ("Program terminated by another instance of TZSet");
                WriteLn;
            <* END *>

            Shutdown := TRUE;
            OS2.DosPostEventSem (MainSem);
        END (*WHILE*);

        OS2.DosResetEventSem (ShutdownSignal, dummy);
        OS2.DosCloseEventSem (ShutdownSignal);

    END TaskKiller;

(************************************************************************)
(*                             MAIN PROGRAM                             *)
(************************************************************************)

PROCEDURE RunTheOperation;

    (* Parses the TZ environment variable, calculates summer time start *)
    (* and end, sets the timezone field in the system DATETIME record.  *)
    (* If there is no summer time, exits.  Otherwise, loops forever,    *)
    (* asleep most of the time, but waking up each time there is a      *)
    (* time zone change.                                                *)

    VAR RegRec: OS2.EXCEPTIONREGISTRATIONRECORD;
        rc: OS2.APIRET;
        delaytime: CARDINAL;

    BEGIN
        (* Initial setting. *)

        GetTZinfo;
        delaytime := MinsToms(DoSetting());
        IF SummerAdjust = 0 THEN
            <* IF SCREENOUT THEN *>
                WriteString ("No further adjustments needed, exiting program.");
                WriteLn;
            <* END *>
        ELSE
            (* Add our break handler to this thread's exception handler chain. *)

            RegRec.prev_structure := NIL;
            RegRec.ExceptionHandler := BreakHandler;
            rc := OS2.DosSetExceptionHandler (RegRec);

            <* IF SCREENOUT THEN *>
                WriteString ("Going to sleep until next time zone update needed.");
                WriteLn;
            <* END *>

            (* Go into an almost-never-ending loop, sleeping until the  *)
            (* next change is due.                                      *)

            WHILE NOT Shutdown DO
                rc := OS2.DosWaitEventSem (MainSem, delaytime);
                IF NOT Shutdown THEN
                    delaytime := TimeToNext();

                    IF delaytime <= 0 THEN

                        (* It is time to switch from winter to summer   *)
                        (* time, or vice versa.                         *)

                        WeAreInSummerTime := NOT WeAreInSummerTime;
                        delaytime := DoSetting();

                    END (*IF*);

                    (* delaytime is in minutes, and we have to convert  *)
                    (* it to milliseconds.  This could cause arithmetic *)
                    (* overflow, so for long delays we will wake up     *)
                    (* early, at which time we will recompute the delay *)
                    (* time.                                            *)

                    delaytime := MinsToms (delaytime);

                END (*IF*);

            END (*WHILE*);

        END (*IF*);

     END RunTheOperation;

(************************************************************************)

(*
PROCEDURE NotDetached(): BOOLEAN;

    (* Returns TRUE if this process is not running detached. *)

    VAR pPib: OS2.PPIB;  pTib: OS2.PTIB;


    BEGIN
        OS2.DosGetInfoBlocks (pTib, pPib);
        RETURN pPib^.pib_ultype <> 4;
    END NotDetached;
*)

(************************************************************************)

VAR tid: OS2.TID;
    rc: OS2.APIRET;

BEGIN
    Shutdown := FALSE;
    <* IF SCREENOUT THEN *>
        WriteString ("TZSet version ");  WriteString (version);  WriteLn;
    <* END *>
    rc := OS2.DosCreateEventSem (NIL, MainSem, 0, FALSE);
    rc := OS2.DosCreateThread (tid, TaskKiller, 0, 0, STACKSIZE);
    RunTheOperation;
FINALLY
    rc := OS2.DosCloseEventSem (MainSem);
END TZSet.

